﻿/* 
* Copyright (c) 2009, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Windows;
using Bio.Data.Providers.Interfaces;
using System;
using System.Windows.Input;
using JulMar.Windows.Mvvm;
using JulMar.Windows.Interfaces;

namespace Bio.Views.ViewModels
{
    /// <summary>
    /// This represents the base class for all BioViewModels
    /// </summary>
    public abstract class BioViewModel : ViewModel
    {
        private string _title;
        public virtual string Title
        {
            get { return _title; }
            set { _title = value; OnPropertyChanged("Title"); }
        }

        public abstract bool Initialize(IBioDataLoader data);
        protected abstract void OnDockStyleChanged();

        public ICommand CloseCommand { get; private set; }

        private bool _isDocked;
        public bool IsDocked
        {
            get { return _isDocked; }
            set 
            {
                _isDocked = value; 
                OnPropertyChanged("IsDocked");
                OnDockStyleChanged();
            }
        }

        protected BioViewModel()
        {
            _isDocked = true;
            CloseCommand = new DelegatingCommand(RaiseCloseRequest);
        }

        protected virtual void OnCreatedView()
        {
        }
    }

    /// <summary>
    /// A UI creator version that manages the connection between the View
    /// and ViewModel and creates the View on the Initialize call.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class BioViewModel<T> : BioViewModel
    {
        public Window CurrentWindow { get; set; }
        public FrameworkElement Visual { get; set; }

        /// <summary>
        /// This initializes the view model and generates the appropriate view.
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public override bool Initialize(IBioDataLoader data)
        {
            // Create the visual
            Visual = (FrameworkElement) Activator.CreateInstance(typeof (T));
            Visual.DataContext = this;

            if (IsDocked == true)
                SendMessage(ViewMessages.NewDockedView, this);
            else
            {
                CurrentWindow = new BioFloatingWindow { DataContext = this };
                CurrentWindow.Closed += win_Closed;
                CurrentWindow.Show();
                CurrentWindow.Activate();
            }

            OnCreatedView();
            return true;
        }

        void win_Closed(object sender, EventArgs e)
        {
            Window win = (Window) sender;
            win.Closed -= win_Closed;
            SendMessage(ViewMessages.Closed, this);
        }

        protected override void OnDockStyleChanged()
        {
            IDisposable waitUI = null;
            var notificationVisualizer = Resolve<INotificationVisualizer>();
            if (notificationVisualizer != null)
                waitUI = notificationVisualizer.BeginWait(string.Empty, string.Empty);

            using (waitUI)
            {
                if (CurrentWindow != null && IsDocked == true)
                {
                    CurrentWindow.Visibility = Visibility.Hidden;
                    CurrentWindow.Content = null;
                    CurrentWindow.Closed -= win_Closed;
                    CurrentWindow.Close();
                    CurrentWindow = null;

                    SendMessage(ViewMessages.NewDockedView, this);
                }
                else
                {
                    SendMessage(ViewMessages.RemoveDockedView, this);
                    CurrentWindow = new BioFloatingWindow {DataContext = this};
                    CurrentWindow.Closed += win_Closed;
                    CurrentWindow.Show();
                }
            }
        }

        public override void RaiseCloseRequest()
        {
            if (IsDocked)
            {
                SendMessage(ViewMessages.RemoveDockedView, this);
                SendMessage(ViewMessages.Closed, this);
            }
            else
            {
                CurrentWindow.Close();
            }

            base.RaiseCloseRequest();
        }
    }
}