﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text.RegularExpressions;
using Bio.Data.Interfaces.Drawing;

namespace Bio.Data.Providers.Structure.XRNA.Internal
{
    internal class XRNABasePairConnectorShape
    {
        public static XRNABasePairConnectorShape Create(string xrnaLine)
        {
            if (xrnaLine != null)
            {
                IBasePairConnectorShape connectorVisual = null;
                BasePairMetadata.ConnectorRenderingMode? style = null;
                string[] ranges = null;
                if (Regex.IsMatch(xrnaLine, BPCircleConnectorDefinition))
                {
                    string[] tokens = Regex.Split(xrnaLine, "#");
                    Debug.Assert(tokens.Length == 2);
                    Debug.Assert(!string.IsNullOrEmpty(tokens[0]));
                    //We need an XRNACircleShape and an Enumerable of positions
                    connectorVisual = RNASecondaryStructureBasePairConnector.FromXRNACircleShape(XRNACircleShape.Create(tokens[1]));
                    style = BasePairMetadata.ConnectorRenderingMode.Circle;
                    ranges = Regex.Split(tokens[0], ",");
                }
                else if (Regex.IsMatch(xrnaLine, BPLineConnectorDefinition))
                {
                    string[] tokens = Regex.Split(xrnaLine, "#");
                    Debug.Assert(tokens.Length == 2);
                    Debug.Assert(!string.IsNullOrEmpty(tokens[0]));
                    //We need an XRNALineShape and an Enumerable of positions
                    connectorVisual = RNASecondaryStructureBasePairConnector.FromXRNALineShape(XRNALineShape.Create(tokens[1]));
                    style = BasePairMetadata.ConnectorRenderingMode.Line;
                    ranges = Regex.Split(tokens[0], ",");
                }

                if (ranges != null && connectorVisual != null)
                {
                    IEnumerable<int> assignedTo = from range in ranges
                                                  from fpIdx in XRNAFile.ExpandXRNARange(range, true)
                                                  select fpIdx - 1;
                    return new XRNABasePairConnectorShape(style, connectorVisual, assignedTo);
                }
            }
            return null; //Some error was encountered or this was not a valid line. In the future, we would probably throw an error to distinguish
        }

        public IEnumerable<int> AssignedTo
        {
            get { return _assignedTo; }
            private set { _assignedTo = value; }
        }

        public IBasePairConnectorShape ConnectorVisual
        {
            get { return _connectorVisual; }
            private set { _connectorVisual = value; }
        }

        public BasePairMetadata.ConnectorRenderingMode? Style
        {
            get { return _style; }
            private set { _style = value; }
        }

        #region Private Members and Methods

        private XRNABasePairConnectorShape(BasePairMetadata.ConnectorRenderingMode? style, IBasePairConnectorShape shapeVisual, IEnumerable<int> assignedTo)
        {
            AssignedTo = assignedTo;
            ConnectorVisual = shapeVisual;
            Style = style;
        }

        private IEnumerable<int> _assignedTo;
        private IBasePairConnectorShape _connectorVisual;
        private BasePairMetadata.ConnectorRenderingMode? _style;
        private static string BPCircleConnectorDefinition = @"[\d+(-?\d+),?]+\s#!Bc\s\d+.\d{2}\s\d+.\d{2}\s\d+.\d{2}\s\d+.\d{2}\s\d\s[a-zA-Z0-9]+";
        private static string BPLineConnectorDefinition = @"[\d+(-?\d+),?]+\s#!Bl\s\d+.\d{2}\s[0-9a-fA-F]+(\s\d){5}";

        #endregion
    }
}
