﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System;
using System.Diagnostics;
using System.Text.RegularExpressions;
using Bio.Data.Interfaces;

namespace Bio.Data.Providers.Structure.XRNA.Internal
{
    internal class XRNABioSequenceVector : IVector
    {
        public static XRNABioSequenceVector Create(string xrnaLine, out int index, out int pairedIndex)
        {
            pairedIndex = -1;
            index = -1;
            if(xrnaLine != null && Regex.IsMatch(xrnaLine, NTDefinition))
            {
                string[] tokens = Regex.Split(xrnaLine, " ");
                Debug.Assert(tokens.Length == 6);
                //Lets do some basic validation
                int proposedIndex, proposedPairedIndex;
                double centerx, centery;
                if (Int32.TryParse(tokens[0], out proposedIndex) &&
                    Int32.TryParse(tokens[5], out proposedPairedIndex) &&
                    Double.TryParse(tokens[2], out centerx) &&
                    Double.TryParse(tokens[3], out centery))
                {
                    if (proposedIndex == proposedPairedIndex) return null; //Error, a nucleotide can't be paired to itself
                    if (proposedIndex < 0 || proposedPairedIndex < 0) return null; //Error, an invalid index is specified
                    pairedIndex = proposedPairedIndex - 1;
                    index = proposedIndex - 1;
                    return new XRNABioSequenceVector(tokens[1][0], centerx, centery);
                }
            }
            return null; //Some error was encountered or this was not a valid line. In the future, we would probably throw an error to distinguish
        }

        #region IVector Members

        public double RayAngle
        {
            get;
            set;
        }

        public double Magnitude
        {
            get;
            set;
        }

        public double StartX
        {
            get;
            set;
        }

        public double StartY
        {
            get;
            set;
        }

        public IVector Clone()
        {
            return new XRNABioSequenceVector(this);
        }

        #endregion

        public void ConnectTo(XRNABioSequenceVector other)
        {
            RayAngle = Math.PI + Math.Atan2((StartY - other.StartY), (StartX - other.StartX));
            Magnitude = Math.Sqrt(Math.Pow((other.StartY - StartY), 2) + Math.Pow((other.StartX - StartX), 2));
        }

        public IBioSymbol BioSymbol
        { 
            get { return _nt; }
            private set { _nt = value; }
        }

        #region Private Members and Methods

        private XRNABioSequenceVector(char symbol, double xpos, double ypos)
        {
            StartX = xpos;
            StartY = -1 * ypos;
            RayAngle = 0.0;
            Magnitude = 0.0;
            BioSymbol = new BioSymbol(BioSymbolType.Nucleotide, symbol);
            ((BioSymbol)BioSymbol).SetExtendedProperty(NucleotideMetadata.Key, new NucleotideMetadata() { Visible = null, Style = null, IsSubstituted = false });
        }

        private XRNABioSequenceVector(XRNABioSequenceVector toCopy)
        {
            this._nt = toCopy._nt;
            this._ntIndex = toCopy._ntIndex;
            this.RayAngle = toCopy.RayAngle;
            this.Magnitude = toCopy.Magnitude;
            this.StartX = toCopy.StartX;
            this.StartY = toCopy.StartY;
        }

        private IBioSymbol _nt;
        private int _ntIndex;
        private static string NTDefinition = @"\d+\s[a-zA-Z]\s-?\d+.\d{2}\s-?\d+.\d{2}\s-?[0-9]2?\s\d+";

        #endregion
    }
}
