﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Windows;
using Bio.Views.Structure.Editor.Internal.Rendering;
using Bio.Views.Structure.Editor.Internal.Rendering.Interfaces;
using Bio.Views.Structure.Editor.Internal.Rendering.Visuals;

namespace Bio.Views.Structure.Editor.Internal.Selection
{
    public class HelixSelectionGroup : StructureElementSelectionGroupBase
    {
        public override int ID { get { return _helixVisual.FivePrimeStartNucleotideIndex; } }
        public override IStructureElementVisualComponent Visual { get { return _helixVisual; } }

        public override void DeleteSelection()
        {
            Manager.RemoveHelix(ID);
        }

        public override void MoveBy(Point delta)
        {
            Manager.MoveHelixBy(ID, delta);
            SaveTranslation(delta);
        }

        public override void MoveTo(Point location)
        {
            Point moveDelta = Manager.MoveHelixTo(ID, location);
            SaveTranslation(moveDelta);
        }

        public override void RotateBy(double angleInDegrees)
        {
            Dictionary<int, Point> edits = Manager.RotateHelixBy(ID, angleInDegrees);
            Manager.SaveStructureElementDrawingOperation(edits);
        }

        public override void RotateTo(double angleInDegrees)
        {
            Dictionary<int, Point> edits = Manager.RotateHelixTo(ID, angleInDegrees);
            Manager.SaveStructureElementDrawingOperation(edits);
        }

        public override Point RotationDelta()
        {
            Rect fivePrimeLocation = Manager.GetLocationForNucleotide(ID);
            return new Point(0 - fivePrimeLocation.Left, 0 - fivePrimeLocation.Top);
        }

        public override void ResizeTo(double newWidth, double newPadding)
        {
            if (newWidth > 0.0)
            {
                Dictionary<int, Point> widthEdits = Manager.ChangeHelixWidth(ID, newWidth);
                Manager.SaveStructureElementDrawingOperation(widthEdits);
            }

            if (newPadding > 0.0)
            {
                Dictionary<int, Point> padEdits = Manager.PadHelix(ID, newPadding);
                Manager.SaveStructureElementDrawingOperation(padEdits);
            }
        }

        public override void SnapHelixToOpeningBasePair()
        {
            Dictionary<int, Point> edits = Manager.SnapHelix(ID, true);
            Manager.SaveStructureElementDrawingOperation(edits);
        }

        public override void SnapHelixToClosingBasePair()
        {
            Dictionary<int, Point> edits = Manager.SnapHelix(ID, false);
            Manager.SaveStructureElementDrawingOperation(edits);
        }

        public HelixSelectionGroup(StructureDrawingManager manager, HelixStructureElementVisualComponent visual)
            : base(manager, visual)
        {
            Debug.Assert(visual != null);
            _helixVisual = visual;
        }

        private void SaveTranslation(Point delta)
        {
            List<int> ntInHelix = new List<int>();
            foreach (var bp in _helixVisual.BasePairs)
            {
                ntInHelix.Add(bp.FivePrimeIndex);
                ntInHelix.Add(bp.ThreePrimeIndex);
            }
            Manager.SaveStructureElementDrawingOperation(ntInHelix, delta);
        }

        private HelixStructureElementVisualComponent _helixVisual;
    }
}
