﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using Bio.Data.Interfaces;
using Bio.Views.Structure.Editor.Internal.Editing;
using Bio.Views.Structure.Editor.Internal.Rendering;
using Bio.Views.Structure.Editor.Internal.Rendering.Interfaces;
using Bio.Views.Structure.Editor.Internal.Rendering.Visuals;

namespace Bio.Views.Structure.Editor.Internal.Selection
{
    public class RangeSelectionGroup : StructureElementSelectionGroupBase
    {
        public override int ID { get { return 0; } }
        public override IStructureElementVisualComponent Visual { get { return _rangeVisual; } }

        public override bool HitTestElement(Point location)
        {
            _lastSelection = _selectableNt.Where(ntVisual => ntVisual.HitTest(location, false)).FirstOrDefault();
            return (_lastSelection != null) ? true : false;
        }

        public override void SelectElement()
        {
            GoToNextStateOnSelect();
        }

        public override void DeselectElement()
        {
            GoToNextStateOnDeselect();
        }

        public override void MoveBy(Point delta)
        {
            if (_currentState == RangeSelectionState.Complete && (_rangeVisual.SelectedNucleotideVisuals.Count() > 0 || _rangeVisual.SelectedBasePairVisuals.Count() > 0))
            {
                //1. Aggregate into one list
                List<int> ntInRange = new List<int>();
                foreach (var ntVisual in _rangeVisual.SelectedNucleotideVisuals)
                    ntInRange.Add(ntVisual.NucleotideIndex);
                foreach (var bpVisual in _rangeVisual.SelectedBasePairVisuals)
                {
                    ntInRange.Add(bpVisual.Data.FivePrimeIndex);
                    ntInRange.Add(bpVisual.Data.ThreePrimeIndex);
                }
                foreach (var ntIdx in ntInRange)
                {
                    Manager.MoveNucleotideBy(ntIdx, delta, true, true, true);
                }
                Manager.SaveStructureElementDrawingOperation(ntInRange, delta);
            }
        }

        public override void EditProperties(StructureElementPropertyChangeSet changeSet, List<IBioSymbol> filter)
        {
            if (_currentState == RangeSelectionState.Complete)
            {
                _rangeVisual.ApplyPropertyEditSet(changeSet, false, filter);
            }
        }

        public override void DeleteSelection()
        {
            //We only need to execute if we have a completed selection.
            if (_currentState == RangeSelectionState.Complete && _rangeVisual.SelectedBasePairVisuals.Count() > 0)
            {
                List<int> localSelectedBasePairs = new List<int>();
                foreach (var sbp in _rangeVisual.SelectedBasePairVisuals)
                    localSelectedBasePairs.Add(sbp.Data.FivePrimeIndex);
                GoToNextStateOnDeselect(); //Clear the selection.
                Manager.RemoveBasePairs(localSelectedBasePairs);
            }
        }

        public override void MakeElementSelectable()
        {
            _selections.Clear();
            _rangeVisual.ClearSelection();
            _currentState = RangeSelectionState.WaitingForFirst;
            _selectableNt = _allNt;
            foreach (var nt in _allNt)
                nt.MakeSelectable();
        }

        public override void MakeElementUnselectable()
        {
            _selections.Clear();
            _rangeVisual.ClearSelection();
            _currentState = RangeSelectionState.WaitingForFirst;
            _selectableNt = Enumerable.Empty<NucleotideStructureElementVisualComponent>();
            foreach (var nt in _allNt)
                nt.MakeUnselectable();
        }

        public RangeSelectionGroup(StructureDrawingManager manager)
            : base(manager, manager.FullModelRangeSelectionVisual)
        {
            Debug.Assert(manager.FullModelRangeSelectionVisual != null);
            _selections = new List<NucleotideStructureElementVisualComponent>();
            _allNt = manager.FullModelRangeSelectionVisual.AllNucleotideVisuals;
            _selectableNt = Enumerable.Empty<NucleotideStructureElementVisualComponent>();
            _currentState = RangeSelectionState.WaitingForFirst;
            _rangeVisual = manager.FullModelRangeSelectionVisual;
        }

        private IEnumerable<NucleotideStructureElementVisualComponent> _selectableNt;
        private IEnumerable<NucleotideStructureElementVisualComponent> _allNt;
        private RangeStructureElementVisualComponent _rangeVisual;
        private RangeSelectionState _currentState;

        private List<NucleotideStructureElementVisualComponent> _selections;
        private NucleotideStructureElementVisualComponent _lastSelection;

        private void GoToNextStateOnSelect()
        {
            if (_currentState == RangeSelectionState.WaitingForFirst)
            {
                if (_lastSelection == null)
                    return; //Were still passive.
                else
                {
                    _selections.Add(_lastSelection);
                    _rangeVisual.SetFirstSelection(_lastSelection);
                    _selectableNt = _allNt.Except(_selections); //Exclude the first selection from the selectable set;
                    _lastSelection = null;
                    _currentState = RangeSelectionState.HaveFirst;
                }   
            }
            else if (_currentState == RangeSelectionState.HaveFirst)
            {
                if (_lastSelection != null)
                {
                    _selections.Add(_lastSelection);
                    _rangeVisual.SetSecondSelection(_lastSelection);
                    _lastSelection = null;
                    _selectableNt = _allNt;
                    _currentState = RangeSelectionState.Complete;
                }
            }
        }

        private void GoToNextStateOnDeselect()
        {
            if (_currentState == RangeSelectionState.HaveFirst)
            {
                //This in effect cancels the range selection becuase the next selection wasn't a second nt.
                if (_lastSelection == null)
                {
                    _rangeVisual.ClearSelection();
                    _selections.Clear();
                    _selectableNt = _allNt;
                    _currentState = RangeSelectionState.WaitingForFirst;
                }
            }
            else if (_currentState == RangeSelectionState.Complete)
            {
                _rangeVisual.ClearSelection();
                _selections.Clear();
                _currentState = RangeSelectionState.WaitingForFirst; //When we come around to the select, if this was the beginning of a new selection, it will start the process over.
            }
        }

        private enum RangeSelectionState { WaitingForFirst, HaveFirst, Complete }
    }
}
