﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using Bio.Views.Structure.Editor.Internal.Selection.Interfaces;

namespace Bio.Views.Structure.Editor.Internal.Tools
{
    public class BasePairCreationTool : SecondaryStructureDrawingTool
    {
        public IEnumerable<IStructureElementSelectionGroup> NucleotideVisuals 
        { 
            get; 
            set; 
        }

        public BasePairCreationTool()
        {
            NucleotideVisuals = Enumerable.Empty<IStructureElementSelectionGroup>();
            Message = null;
            Icon = "B";
            ToolTip = "Base Pair Creation Tool; Create new Base Pair by sequentially selecting the 5' and 3' Nt.";
        }

        protected override void OnActiveStateChanged()
        {
            if (!IsActive)
                Message = null;
            else
            {
                FivePrimeNtSelection = null;
                ThreePrimeNtSelection = null;
                State = BasePairCreationState.WaitingForFivePrimeNt;
                SetMessage();
            }
        }

        protected override void DrawingPreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (IsActive)
            {
                Point location = e.GetPosition(Drawing);
                var selectedNucleotide = NucleotideVisuals.Where(ntVisual => ntVisual.HitTestElement(location)).FirstOrDefault();
                GoToNextState(selectedNucleotide);
            }
        }

        private IStructureElementSelectionGroup FivePrimeNtSelection { get; set; }
        private IStructureElementSelectionGroup ThreePrimeNtSelection { get; set; }
        private BasePairCreationState State { get; set; }
        private enum BasePairCreationState { WaitingForFivePrimeNt, WaitingForThreePrimeNt, Complete }
        private static string BASE_PAIR_CREATION_TOOL_TITLE = "Pair Creator";

        private void SetMessage()
        {
            switch (State)
            {
                case BasePairCreationState.WaitingForFivePrimeNt:
                case BasePairCreationState.Complete:
                    Message = string.Format("{0}:\nSelect the 5' NT", BASE_PAIR_CREATION_TOOL_TITLE);
                    break;
                case BasePairCreationState.WaitingForThreePrimeNt:
                    Message = string.Format("{0}:\nSelect a 3' NT\n 5' NT: {1}", BASE_PAIR_CREATION_TOOL_TITLE, FivePrimeNtSelection.ID + 1);
                    break;
                default:
                    Message = string.Format("{0}:\nSelect the 5' NT", BASE_PAIR_CREATION_TOOL_TITLE);
                    break;
            };
        }

        private void GoToNextState(IStructureElementSelectionGroup selection)
        {
            if (State == BasePairCreationState.WaitingForFivePrimeNt)
            {
                if (selection != null)
                {
                    FivePrimeNtSelection = selection;
                    ThreePrimeNtSelection = null;
                    State = BasePairCreationState.WaitingForThreePrimeNt;
                    SetMessage();   
                }
            }
            else if (State == BasePairCreationState.WaitingForThreePrimeNt)
            {
                if (FivePrimeNtSelection != null)
                {
                    if (selection != null)
                    {
                        ThreePrimeNtSelection = selection;
                        Drawing.CreateBasePair(FivePrimeNtSelection.ID, ThreePrimeNtSelection.ID);
                        State = BasePairCreationState.Complete;
                    }
                    else
                    {
                        FivePrimeNtSelection = null;
                        State = BasePairCreationState.WaitingForFivePrimeNt;
                        Message = string.Empty;
                    }
                }
                else
                {
                    //Some kind of error state.
                    FivePrimeNtSelection = null;
                    State = BasePairCreationState.WaitingForFivePrimeNt;
                }
                SetMessage();
            }
            else if (State == BasePairCreationState.Complete)
            {
                if (selection == null)
                {
                    FivePrimeNtSelection = null;
                    ThreePrimeNtSelection = null;
                    State = BasePairCreationState.WaitingForFivePrimeNt;
                }
                else
                {
                    FivePrimeNtSelection = selection;
                    ThreePrimeNtSelection = null;
                    State = BasePairCreationState.WaitingForThreePrimeNt;
                }
                SetMessage();
            }
        }
    }
}
