﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using Bio.Views.Structure.Editor.Internal.Selection.Interfaces;

namespace Bio.Views.Structure.Editor.Internal.Tools
{
    public class HelixCreationTool : SecondaryStructureDrawingTool
    {
        public IEnumerable<IStructureElementSelectionGroup> NucleotideVisuals
        {
            get;
            set;
        }

        /// <summary>
        /// Basically a three-way state machine to select:
        /// 1. FivePrimeNtStart
        /// 2. ThreePrimeNtEnd
        /// 3. FivePrimeNtEnd
        /// 
        /// The ThreePrimeNtStart can then be inferred.
        /// </summary>
        public HelixCreationTool()
        {
            NucleotideVisuals = Enumerable.Empty<IStructureElementSelectionGroup>();
            Message = null;
            Icon = "H";
            ToolTip = "Helix Creation Tool; Create a new Helix by sequentially selecting the 5' start Nt, 3' end Nt and 5' end Nt";
        }

        protected override void OnActiveStateChanged()
        {
            if (!IsActive)
                Message = null;
            else
            {
                FivePrimeNtStartSelection = null;
                ThreePrimeNtEndSelection = null;
                FivePrimeNtEndSelection = null;
                State = HelixCreationState.WaitingForFivePrimeStart;
                SetMessage();
            }
        }

        protected override void DrawingPreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (IsActive)
            {
                Point location = e.GetPosition(Drawing);
                var selectedNucleotide = NucleotideVisuals.Where(ntVisual => ntVisual.HitTestElement(location)).FirstOrDefault();
                GoToNextState(selectedNucleotide);
            }
        }

        private IStructureElementSelectionGroup FivePrimeNtStartSelection { get; set; }
        private IStructureElementSelectionGroup ThreePrimeNtEndSelection { get; set; }
        private IStructureElementSelectionGroup FivePrimeNtEndSelection { get; set; }
        private enum HelixCreationState { WaitingForFivePrimeStart, WaitingForThreePrimeEnd, WaitingForFivePrimeEnd, Complete }
        private HelixCreationState State { get; set; }
        private static string HELIX_CREATION_TOOL_TITLE = "Helix Creator";

        private void SetMessage()
        {
            switch (State)
            {
                case HelixCreationState.WaitingForFivePrimeStart:
                case HelixCreationState.Complete:
                    Message = string.Format("{0}:\nSelect the 5' Start NT", HELIX_CREATION_TOOL_TITLE);
                    break;
                case HelixCreationState.WaitingForThreePrimeEnd:
                    Message = string.Format("{0}:\nSelect a 3' End NT\n 5' Start NT: {1}", HELIX_CREATION_TOOL_TITLE, FivePrimeNtStartSelection.ID + 1);
                    break;
                case HelixCreationState.WaitingForFivePrimeEnd:
                    Message = string.Format("{0}:\nSelect the 5' End NT\n 5' Start NT: {1}\n 3' End Nt: {2}", HELIX_CREATION_TOOL_TITLE, FivePrimeNtStartSelection.ID + 1, ThreePrimeNtEndSelection.ID + 1);
                    break;
                default:
                    Message = string.Format("{0}:\nSelect the 5' Start NT", HELIX_CREATION_TOOL_TITLE);
                    break;
            };
        }

        private void GoToNextState(IStructureElementSelectionGroup selection)
        {
            if (State == HelixCreationState.WaitingForFivePrimeStart)
            {
                if (selection != null)
                {
                    FivePrimeNtStartSelection = selection;
                    ThreePrimeNtEndSelection = null;
                    FivePrimeNtEndSelection = null;
                    State = HelixCreationState.WaitingForThreePrimeEnd;
                    SetMessage();
                }
            }
            else if (State == HelixCreationState.WaitingForThreePrimeEnd)
            {
                if (FivePrimeNtStartSelection != null)
                {
                    if (selection != null)
                    {
                        ThreePrimeNtEndSelection = selection;
                        FivePrimeNtEndSelection = null;
                        State = HelixCreationState.WaitingForFivePrimeEnd;
                    }
                    else
                    {
                        FivePrimeNtStartSelection = null;
                        ThreePrimeNtEndSelection = null;
                        FivePrimeNtEndSelection = null;
                        State = HelixCreationState.WaitingForFivePrimeStart;
                    }
                }
                else
                {
                    //Some kind of error state.
                    FivePrimeNtStartSelection = null;
                    ThreePrimeNtEndSelection = null;
                    FivePrimeNtEndSelection = null;
                    State = HelixCreationState.WaitingForFivePrimeStart;
                }
                SetMessage();
            }
            else if(State == HelixCreationState.WaitingForFivePrimeEnd)
            {
                if(ThreePrimeNtEndSelection != null && FivePrimeNtStartSelection != null)
                {
                    if(selection != null)
                    {
                        FivePrimeNtEndSelection = selection;
                        Drawing.CreateHelix(FivePrimeNtStartSelection.ID, ThreePrimeNtEndSelection.ID, FivePrimeNtEndSelection.ID);
                        State = HelixCreationState.Complete;
                    }
                    else
                    {
                        FivePrimeNtStartSelection = null;
                        ThreePrimeNtEndSelection = null;
                        FivePrimeNtEndSelection = null;
                        State = HelixCreationState.WaitingForFivePrimeStart;   
                    }
                    SetMessage();
                }
                else
                {
                    //Some kind of error state.
                    FivePrimeNtStartSelection = null;
                    ThreePrimeNtEndSelection = null;
                    FivePrimeNtEndSelection = null;
                    State = HelixCreationState.WaitingForFivePrimeStart;
                }
                SetMessage();
            }
            else if (State == HelixCreationState.Complete)
            {
                if (selection == null)
                {
                    FivePrimeNtStartSelection = null;
                    ThreePrimeNtEndSelection = null;
                    FivePrimeNtEndSelection = null;
                    State = HelixCreationState.WaitingForFivePrimeStart;
                }
                else
                {
                    FivePrimeNtStartSelection = selection;
                    ThreePrimeNtEndSelection = null;
                    FivePrimeNtEndSelection = null;
                    State = HelixCreationState.WaitingForThreePrimeEnd;
                }
                SetMessage();
            }
        }
    }
}
