﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Windows;
using System.Windows.Input;
using Bio.Views.Structure.Editor.Internal.Selection;

namespace Bio.Views.Structure.Editor.Internal.Tools
{
    public class StructureElementRotateTool : SecondaryStructureDrawingTool
    {
        public StructureElementRotateTool()
        {
            Icon = "R";
            Message = null;
            ToolTip = "Rotation Tool; Rotate a Base Pair about its 5' Nt or a Helix about its 5' start Nt.";
        }

        public override void OnSelectionModeChanged()
        {
            if(IsActive)
                SetMessage();
        }

        protected override void DrawingMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            if(Drawing.TryStartDragRotation(clickLocation))
                GoToNextState();
        }

        protected override void DrawingMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            double rotation;
            if(Drawing.TryEndDragRotation(clickLocation, out rotation))
                GoToNextState();
        }

        protected override void DrawingMouseMove(object sender, MouseEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            double rotation;
            if (Drawing.TryUpdateDragRotation(clickLocation, out rotation))
            {
                RotationAngle = rotation;
                SetMessage();
            }
        }

        protected override void DrawingMouseLeave(object sender, MouseEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            double rotation;
            if(Drawing.TryEndDragRotation(clickLocation, out rotation))
                GoToNextState();
        }

        protected override void OnActiveStateChanged()
        {
            if (!IsActive)
                Message = null;
            else
            {
                State = RotationState.Waiting;
                SetMessage();
            }
        }

        private RotationState State { get; set; }
        private enum RotationState { Waiting, InRotation, Complete };
        private static string ROTATION_TOOL_TITLE = "Rotation Tool";
        private double RotationAngle { get; set; }
        private double LastRotationAngle { get; set; }

        private void SetMessage()
        {
            switch (State)
            {
                case RotationState.Waiting:
                    if (Drawing.SelectionMode == StructureElementSelectionMode.None)
                    {
                        Message = string.Format("{0}: Must set the Selection Mode to something other than None to use this Tool", ROTATION_TOOL_TITLE);
                    }
                    else
                    {
                        Message = ROTATION_TOOL_TITLE;
                    }
                    break;
                case RotationState.InRotation:
                    Message = string.Format("{0}: Rotation: {1:F} degrees", ROTATION_TOOL_TITLE, RotationAngle);
                    break;
                case RotationState.Complete:
                    Message = string.Format("{0}: Last Rotation: {1:F} degrees", ROTATION_TOOL_TITLE, LastRotationAngle);
                    break;
                default:
                    Message = ROTATION_TOOL_TITLE;
                    break;
            };
        }

        private void GoToNextState()
        {
            if (State == RotationState.Waiting)
                State = RotationState.InRotation;
            else if (State == RotationState.InRotation)
            {
                State = RotationState.Complete;
                LastRotationAngle = RotationAngle;
                RotationAngle = 0.0;
            }
            else if (State == RotationState.Complete)
                State = RotationState.InRotation;
            SetMessage();
        }
    }
}
