﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows.Input;
using Bio.Data;
using Bio.Data.Interfaces;
using Bio.Views.Structure.Editor.Internal.Data;
using Bio.Views.Structure.Editor.Views;
using JulMar.Windows;
using JulMar.Windows.Interfaces;
using JulMar.Windows.Mvvm;

namespace Bio.Views.Structure.Editor.ViewModels
{
    public class NucleotideVisualRenderingDefaultsViewModel : ViewModel
    {
        public NucleotideVisualRenderingOptionsViewModel SelectedOptionsSet
        {
            get { return _selectedOptionsSet; }
            set
            {
                _selectedOptionsSet = value;
                OnPropertyChanged("SelectedOptionsSet");
            }
        }

        public MTObservableCollection<NucleotideVisualRenderingOptionsViewModel> OptionsSets
        {
            get { return _allOptions; }
            private set
            {
                _allOptions = value;
                OnPropertyChanged("OptionsSets");
            }
        }

        public NucleotideVisualRenderingOptionsDataCollection RuntimeOptionsSets
        {
            get;
            private set;
        }

        public ICommand AddNewNucleotideRenderingDefaultsOptionSetCommand { get; private set; }
        public ICommand DeleteNucleotideRenderingDefaultsOptionSetCommand { get; private set; }
        public ICommand ResetRuntimeDefaultNucleotideRenderingOptionsToBuiltInDefaultsCommand { get; private set; }

        public NucleotideVisualRenderingDefaultsViewModel(NucleotideVisualRenderingOptionsDataCollection runtimeOptionsSets, NucleotideVisualRenderingOptionsDataCollection defaultOptionsSets)
        {
            Debug.Assert(defaultOptionsSets != null);
            Debug.Assert(runtimeOptionsSets != null);
            NucleotideVisualRenderingOptionsDataCollection localRuntimeOptionsSets = new NucleotideVisualRenderingOptionsDataCollection();
            foreach (var optionSet in runtimeOptionsSets)
                localRuntimeOptionsSets.Add(new NucleotideVisualRenderingOptionsData(optionSet));
            RuntimeOptionsSets = localRuntimeOptionsSets;
            _defaultOptionsSet = defaultOptionsSets;
            Initialize();
        }

        private void Initialize()
        {
            AddNewNucleotideRenderingDefaultsOptionSetCommand = new DelegatingCommand(OnAddNewNucleotideDefaultOptionSet);
            ResetRuntimeDefaultNucleotideRenderingOptionsToBuiltInDefaultsCommand = new DelegatingCommand(OnResetRuntimeDefaultNucleotideOptionSets);
            OptionsSets = new MTObservableCollection<NucleotideVisualRenderingOptionsViewModel>();
            foreach (var optionSet in RuntimeOptionsSets)
            {
                OptionsSets.Add(new NucleotideVisualRenderingOptionsViewModel(optionSet));
            }

            SelectedOptionsSet = OptionsSets[0];

            _validNucleotides = new List<IBioSymbol>()
            {
                new BioSymbol(BioSymbolType.Nucleotide, 'A'),
                new BioSymbol(BioSymbolType.Nucleotide, 'G'),
                new BioSymbol(BioSymbolType.Nucleotide, 'C'),
                new BioSymbol(BioSymbolType.Nucleotide, 'U'),
                new BioSymbol(BioSymbolType.Nucleotide, 'T'),
                new BioSymbol(BioSymbolType.Nucleotide, 'R'),
                new BioSymbol(BioSymbolType.Nucleotide, 'Y'),
                new BioSymbol(BioSymbolType.Nucleotide, 'N')
            };
        }

        private void OnAddNewNucleotideDefaultOptionSet()
        {
            //We will have to put up a simple UI to solicit a label
            SelectNucleotideForRenderingDefaultsViewModel vm = new SelectNucleotideForRenderingDefaultsViewModel(_validNucleotides);
            IUIVisualizer uiVisualizer = ViewModel.ServiceProvider.Resolve<IUIVisualizer>();
            if (uiVisualizer != null)
            {
                bool? result = uiVisualizer.ShowDialog(typeof(SelectNucleotideForRenderingDefaultsEditor).ToString(), vm);
                if (result != null && result.Value && vm.SelectedValue != null)
                {
                    var existingOptionSet = OptionsSets.Where(set => set.Label.Equals(vm.SelectedValue.Value.ToString())).FirstOrDefault();
                    if (existingOptionSet != null)
                    {
                        IErrorVisualizer errorVisualizer = Resolve<IErrorVisualizer>();
                        if (errorVisualizer != null)
                        {
                            errorVisualizer.Show("Error", string.Format("Error: Already have an Options Set for Nucleotide: {0}", vm.SelectedValue.Value.ToString()));
                        }
                    }
                    else
                    {
                        NucleotideVisualRenderingOptionsData newOptionsSet = NucleotideVisualRenderingOptionsData.CreateDefault(vm.SelectedValue.Value.ToString());
                        NucleotideVisualRenderingOptionsViewModel newOptionsSetVM = new NucleotideVisualRenderingOptionsViewModel(newOptionsSet);
                        OptionsSets.Add(newOptionsSetVM);
                        RuntimeOptionsSets.Add(newOptionsSet);
                        SelectedOptionsSet = newOptionsSetVM; //Make this new set selected.
                    }
                }
            }
        }

        private void OnResetRuntimeDefaultNucleotideOptionSets()
        {
            //We'll just replace the runtime with the default.
            RuntimeOptionsSets.Clear();
            foreach (var defaultOptionSet in _defaultOptionsSet)
                RuntimeOptionsSets.Add(new NucleotideVisualRenderingOptionsData(defaultOptionSet));
            OptionsSets.Clear();
            foreach (var runtimeOptionSet in RuntimeOptionsSets)
                OptionsSets.Add(new NucleotideVisualRenderingOptionsViewModel(runtimeOptionSet));
            SelectedOptionsSet = OptionsSets[0];
        }

        private MTObservableCollection<NucleotideVisualRenderingOptionsViewModel> _allOptions;
        private List<IBioSymbol> _validNucleotides;
        private NucleotideVisualRenderingOptionsViewModel _selectedOptionsSet;
        private NucleotideVisualRenderingOptionsDataCollection _defaultOptionsSet;
    }
}
