﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data.SqlClient;
using System.Linq;
using System.Threading;
using System.Windows;
using System.Windows.Input;
using Bio.Data.Providers.rCAD.Structure.Models;
using JulMar.Windows;
using JulMar.Windows.Behaviors;
using JulMar.Windows.Extensions;
using JulMar.Windows.Mvvm;
using JulMar.Windows.Validations;

namespace Bio.Data.Providers.rCAD.Structure.ViewModels
{
    public class SetViewModel  : ValidatingViewModel
    {
        private int _step, _matchingSequencesWithStructureModels = -1;
        private TaxonomyEntry _rootTaxonomy;
        private readonly StructureModelSet _structureModels;

        /// <summary>
        /// User friendly name for the filter
        /// </summary>
        [Required]
        public string Name
        {
            get { return _structureModels.Name; }
            set { _structureModels.Name = value; OnPropertyChanged("Name"); }
        }

        public bool LimitToStructureModelsWithRendering
        {
            get { return _structureModels.LimitToModelsWith2DRendering; }
            set
            {
                _structureModels.LimitToModelsWith2DRendering = value;
                _matchingSequencesWithStructureModels = -1; //We'll need to recompute the filter
                ClearSequencesWithStructureModelsSelectionList(); 
                _structureModels.SelectedSet.Clear();
                SelectAllModelsInList = false;
                OnPropertyChanged("LimitToStructureModelsWithRendering", "MatchingSequenceCount", "HasValidSequenceList", "SequencesWithStructureModelsSelectionList");
            }
        }

        /// <summary>
        /// Root taxononmy ID used for filter.  Only sequences under this ID will be retrieved.
        /// </summary>
        public int ParentTaxId
        {
            get { return _structureModels.ParentTaxId; }
            set
            {
                _structureModels.ParentTaxId = value;
                _matchingSequencesWithStructureModels = -1;
                OnPropertyChanged("ParentTaxId", "IsValid", "MatchingSequenceCount", "HasValidSequenceList");
            }
        }

        /// <summary>
        /// RNA location identifier
        /// </summary>
        public int LocationId
        {
            get { return _structureModels.LocationId; }
            set
            {
                _structureModels.LocationId = value;
                _matchingSequencesWithStructureModels = -1;
                ClearSequencesWithStructureModelsSelectionList();
                _structureModels.SelectedSet.Clear();
                OnPropertyChanged("LocationId", "MatchingSequenceCount", "HasValidSequenceList", "SequencesWithStructureModelsSelectionList");
            }
        }

        /// <summary>
        /// Sequence type (5s, 16s, 23s)
        /// </summary>
        public int SequenceTypeId
        {
            get { return _structureModels.SequenceTypeId; }
            set
            {
                _structureModels.SequenceTypeId = value;
                _matchingSequencesWithStructureModels = -1;
                _alignments = null;
                ClearSequencesWithStructureModelsSelectionList();
                _structureModels.SelectedSet.Clear();
                OnPropertyChanged("SequenceTypeId", "MatchingSequenceCount", "HasValidSequenceList", "Alignments", "SequencesWithStructureModelsSelectionList");
            }
        }

        /// <summary>
        /// Alignment id
        /// </summary>
        public int AlignmentId
        {
            get { return _structureModels.AlignmentId; }
            set
            {
                _structureModels.AlignmentId = value;
                _matchingSequencesWithStructureModels = -1;
                ClearSequencesWithStructureModelsSelectionList();
                _structureModels.SelectedSet.Clear();
                OnPropertyChanged("AlignmentId", "MatchingSequenceCount", "HasValidSequenceList", "SequencesWithStructureModelsSelectionList");
            }
        }

        /// <summary>
        /// The server name (.\SQLEXPRESS, (local), etc.)
        /// </summary>
        [Required]
        public string Server
        {
            get { return Connection.Server; }
            set
            {
                Connection.Server = value;
                OnPropertyChanged("Server");
                ResetConnectionDependencies();
                ThreadPool.QueueUserWorkItem(o => OnFillDatabases(false));
            }
        }

        /// <summary>
        /// Returns the total count of matching sequences found.
        /// </summary>
        public int MatchingSequenceCount
        {
            get
            {
                if (IsValid && _matchingSequencesWithStructureModels == -1 && !SelectSpecificStructureModels)
                {
                    using (var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString))
                    {
                        const string sqlQuery = @"with cte as (select TaxID from Taxonomy where TaxID={0} " +
                                                @"union all select t.TaxID from Taxonomy t join cte tp on t.ParentTaxID = tp.TaxID) " +
                                                @"select * from cte";
                        var results = dc
                            .ExecuteQuery<Taxonomy>(sqlQuery, ParentTaxId)
                            .Select(tax => tax.TaxID).ToList();

                        var queryGroups =
                            results.Select((id, index) => new { GroupID = index / 2000, id }).GroupBy(x => x.GroupID);

                        _matchingSequencesWithStructureModels = 0;
                        _structureModels.SelectedSet.Clear();

                        //I took this from the rCAD Alignment Provider, but modified the query to check if the sequence has any structure models.
                        foreach (var singleGroup in queryGroups)
                        {
                            var thisGroup = singleGroup;
                            //KJD (10/21/2009) - Change query to use vAlignmentGridUngapped which takes into account column indirection supported by the rCAD schema
                            //instead of querying the Sequence table directly. We also now enforce that a specific alignment is used to determine sequence counts.
                            if (LimitToStructureModelsWithRendering)
                            {
                                var ids = (from seq in dc.SequenceMains
                                           where thisGroup.Select(x => x.id).Contains(seq.TaxID)
                                                 && (LocationId == 0 || seq.LocationID == LocationId)
                                                 && (seq.SeqTypeID == SequenceTypeId)
                                                 && dc.SecondaryStructureDrawings.Where(s => s.SeqID == seq.SeqID && s.AlnID == AlignmentId).FirstOrDefault() != null
                                                 && dc.vAlignmentGridUngappeds.Where(s => s.SeqID == seq.SeqID && s.AlnID == AlignmentId && s.LogicalColumnNumber > 0).FirstOrDefault() != null
                                           select seq.SeqID);
                                _structureModels.SelectedSet.AddRange(ids);
                            }
                            else
                            {
                                var ids = (from seq in dc.SequenceMains
                                           where thisGroup.Select(x => x.id).Contains(seq.TaxID)
                                                 && (LocationId == 0 || seq.LocationID == LocationId)
                                                 && (seq.SeqTypeID == SequenceTypeId)
                                                 && dc.SecondaryStructureBasePairs.Where(s => s.SeqID == seq.SeqID && s.AlnID == AlignmentId).FirstOrDefault() != null
                                                 && dc.vAlignmentGridUngappeds.Where(s => s.SeqID == seq.SeqID && s.AlnID == AlignmentId && s.LogicalColumnNumber > 0).FirstOrDefault() != null
                                           select seq.SeqID);
                                _structureModels.SelectedSet.AddRange(ids);
                            }

                            _matchingSequencesWithStructureModels = _structureModels.SelectedSet.Count;
                        }

                        OnPropertyChanged("HasValidSequenceList");

                        return _matchingSequencesWithStructureModels;
                    }
                }
                else
                {
                    if (SelectSpecificStructureModels)
                    {
                        return _matchingSequencesWithStructureModels;
                    }
                    else
                    {
                        return 0;
                    }

                }
            }
        }

        /// <summary>
        /// Provides limitations on the sequence list.
        /// </summary>
        public bool HasValidSequenceList
        {
            get { return (_matchingSequencesWithStructureModels > 0 && _matchingSequencesWithStructureModels < 2000); }
        }

        private void ResetConnectionDependencies()
        {
            _rootTaxonomy = null;
            _seqTypes = null;
            _locationTypes = null;
            _matchingSequencesWithStructureModels = -1;
            ClearSequencesWithStructureModelsSelectionList();
            _structureModels.SelectedSet.Clear();
            OnPropertyChanged("IsValid", "ConnectionString", "MatchingSequenceCount",
                "TaxonomyRoot", "SequenceTypes", "CellLocationTypes", "SequencesWithStructureModelsSelectionList");
        }

        /// <summary>
        /// The database name on above server
        /// </summary>
        [Required]
        public string Database
        {
            get { return Connection.Database; }
            set
            {
                Connection.Database = value;
                OnPropertyChanged("Database");
                ResetConnectionDependencies();
            }
        }

        /// <summary>
        /// Returns if this is SQL security
        /// </summary>
        public bool IsSqlServerSecurity
        {
            get { return Connection.SecurityType == SecurityType.SqlAuthentication; }
            set
            {
                Connection.SecurityType = (value == true) ? SecurityType.SqlAuthentication : SecurityType.WindowsAuthentication;
                OnPropertyChanged("IsSqlServerSecurity", "IsWindowsSecurity");
                ResetConnectionDependencies();
            }
        }

        /// <summary>
        /// Returns if this is Windows security
        /// </summary>
        public bool IsWindowsSecurity
        {
            get { return Connection.SecurityType != SecurityType.SqlAuthentication; }
            set
            {
                Connection.SecurityType = value == false ? SecurityType.SqlAuthentication : SecurityType.WindowsAuthentication;
                OnPropertyChanged("IsSqlServerSecurity", "IsWindowsSecurity");
                ResetConnectionDependencies();
                ThreadPool.QueueUserWorkItem(o => OnFillDatabases(false));
            }
        }

        /// <summary>
        /// User name (if SecurityType = SQL)
        /// </summary>
        public string Username
        {
            get { return Connection.Username; }
            set
            {
                Connection.Username = value;
                OnPropertyChanged("Username");
                ResetConnectionDependencies();
            }
        }

        /// <summary>
        /// Password for SecurityType = SQL
        /// </summary>
        public string Password
        {
            get { return Connection.Password; }
            set
            {
                Connection.Password = value;
                OnPropertyChanged("Password");
                ResetConnectionDependencies();
            }
        }

        /// <summary>
        /// True to serialize connection
        /// </summary>
        private bool _shouldSerialize;
        public bool ShouldSerialize
        {
            get { return _shouldSerialize; }
            set { _shouldSerialize = value; OnPropertyChanged("ShouldSerialize"); }
        }

        /// <summary>
        /// Returns whether this connection is "valid"
        /// </summary>
        public bool IsValid
        {
            get { return _structureModels.IsValid; }
        }

        /// <summary>
        /// Returns the connection string
        /// </summary>
        public string ConnectionString
        {
            get { return Connection.BuildConnectionString(); }
        }

        /// <summary>
        /// Retrieve the Taxonomy information from the database.
        /// </summary>
        public IEnumerable<TaxonomyEntry> TaxonomyRoot
        {
            get
            {
                if (_rootTaxonomy == null && Connection.IsValid)
                    _rootTaxonomy = new TaxonomyEntry(RcadSecondaryStructureProvider.CreateDbContext(ConnectionString), true);
                yield return _rootTaxonomy;
            }
        }

        private bool _selectSpecificStructureModels;
        public bool SelectSpecificStructureModels
        {
            get { return _selectSpecificStructureModels; }
            set
            {
                _selectSpecificStructureModels = value;
                _matchingSequencesWithStructureModels = -1; //We'll need to recompute the filter
                ClearSequencesWithStructureModelsSelectionList();
                _structureModels.SelectedSet.Clear();
                SelectAllModelsInList = false;
                OnPropertyChanged("SelectSpecificStructureModels", "SequencesWithStructureModelsSelectionList");
            }
        }

        private bool _selectAllModelsInList;
        public bool SelectAllModelsInList
        {
            get { return _selectAllModelsInList; }
            set
            {
                _selectAllModelsInList = value;
                if (SelectSpecificStructureModels && SequencesWithStructureModelsSelectionList!=null)
                {
                    foreach (var entry in SequencesWithStructureModelsSelectionList)
                    {
                        entry.IsSelected = SelectAllModelsInList;
                    }
                }
                OnPropertyChanged("SelectAllModelsInList");
            }
        }

        private void ClearSequencesWithStructureModelsSelectionList()
        {
            if (_sequencesWithStructureModelsSelectionList != null)
            {
                foreach (var entry in _sequencesWithStructureModelsSelectionList)
                {
                    entry.SelectionStateChanged -= new SequenceEntry.SelectionStateChangedEventHandler(OnStructureModelSelectedFromList);
                }
                _sequencesWithStructureModelsSelectionList = null;
            }
        }

        private ObservableCollection<SequenceEntry> _sequencesWithStructureModelsSelectionList;
        public ObservableCollection<SequenceEntry> SequencesWithStructureModelsSelectionList
        {
            get
            {
                if (_sequencesWithStructureModelsSelectionList == null && Connection.IsValid && SelectSpecificStructureModels)
                {
                    _sequencesWithStructureModelsSelectionList = new ObservableCollection<SequenceEntry>();
                    _structureModels.SelectedSet.Clear();
                    using (var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString))
                    {
                        if (LimitToStructureModelsWithRendering)
                        {
                            var seqWithSMSet = (from seq in dc.SequenceMains
                                                join seqinaln in dc.vAlignmentGridUngappeds
                                                on seq.SeqID equals seqinaln.SeqID
                                                join strmodels in dc.SecondaryStructureDrawings
                                                on seq.SeqID equals strmodels.SeqID
                                                join taxname in dc.TaxonomyNames
                                                on seq.TaxID equals taxname.TaxID
                                                where seq.SeqTypeID == _structureModels.SequenceTypeId &&
                                                      (_structureModels.LocationId == 0 || seq.LocationID == _structureModels.LocationId) &&
                                                      seqinaln.AlnID == _structureModels.AlignmentId
                                                select new { seq.SeqID, taxname.ScientificName }).Distinct();

                            foreach (var seqWithSM in seqWithSMSet)
                            {
                                SequenceEntry entry = new SequenceEntry(_structureModels)
                                {
                                    Id = seqWithSM.SeqID,
                                    IsSelected = false,
                                    Name = seqWithSM.ScientificName,
                                    HasRendering = true
                                };
                                entry.SelectionStateChanged += new SequenceEntry.SelectionStateChangedEventHandler(OnStructureModelSelectedFromList);
                                _sequencesWithStructureModelsSelectionList.Add(entry);
                            }
                        }
                        else
                        {
                            var seqWithSMSetRI = (from seq in dc.SequenceMains
                                                  join seqinaln in dc.vAlignmentGridUngappeds
                                                  on seq.SeqID equals seqinaln.SeqID
                                                  join strmodels in dc.SecondaryStructureBasePairs.Select(x => new { x.SeqID, x.AlnID }).Distinct()
                                                  on seq.SeqID equals strmodels.SeqID
                                                  join taxname in dc.TaxonomyNames
                                                  on seq.TaxID equals taxname.TaxID
                                                  let hasRendering = dc.SecondaryStructureDrawings.Where(x => ((x.SeqID == seq.SeqID) && (x.AlnID == _structureModels.AlignmentId))).FirstOrDefault() != null
                                                  where seq.SeqTypeID == _structureModels.SequenceTypeId &&
                                                                         (_structureModels.LocationId == 0 || seq.LocationID == _structureModels.LocationId) &&
                                                                          seqinaln.AlnID == _structureModels.AlignmentId
                                                  select new { seq.SeqID, hasRendering, taxname.ScientificName }).Distinct();

                            foreach (var seqWithSMRI in seqWithSMSetRI)
                            {
                                SequenceEntry entry = new SequenceEntry(_structureModels)
                                {
                                    Id = seqWithSMRI.SeqID,
                                    HasRendering = seqWithSMRI.hasRendering,
                                    Name = seqWithSMRI.ScientificName,
                                    IsSelected = false
                                };
                                entry.SelectionStateChanged += new SequenceEntry.SelectionStateChangedEventHandler(OnStructureModelSelectedFromList);
                                _sequencesWithStructureModelsSelectionList.Add(entry);
                            }
                                                                                
                        }
                    }
                }
                return _sequencesWithStructureModelsSelectionList;
            }
        }

        // Retrieve the RNA based sequence types
        private MTObservableCollection<KeyNameData> _seqTypes;
        public MTObservableCollection<KeyNameData> SequenceTypes
        {
            get
            {
                if (_seqTypes == null && Connection.IsValid)
                {
                    try
                    {
                        _seqTypes = new MTObservableCollection<KeyNameData>();

                        using (var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString))
                        {
                            _seqTypes.AddRange(from st in dc.SequenceTypes
                                               where st.MoleculeType == "RNA"
                                               orderby st.GeneType
                                               select new KeyNameData
                                               {
                                                   Id = st.SeqTypeID,
                                                   Name = st.GeneType + " - " + st.GeneName
                                               });
                        }

                        if (_structureModels.SequenceTypeId < 1 && _seqTypes.Count > 0)
                        {
                            var entry = (_seqTypes.FirstOrDefault(s => s.Name.StartsWith("Ribo")));
                            SequenceTypeId = (entry != null) ? entry.Id : _seqTypes[0].Id;
                        }
                    }
                    catch
                    {
                        _seqTypes = null;
                    }
                }
                return _seqTypes;
            }
        }

        // Retrieve the call location types
        private ObservableCollection<KeyNameData> _locationTypes;
        public ObservableCollection<KeyNameData> CellLocationTypes
        {
            get
            {
                if (_locationTypes == null && Connection.IsValid)
                {
                    try
                    {
                        _locationTypes = new ObservableCollection<KeyNameData> { new KeyNameData { Id = 0, Name = "All Cell Locations" } };

                        using (var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString))
                        {
                            _locationTypes.AddRange(from cl in dc.CellLocationInfos
                                                    orderby cl.Description
                                                    select new KeyNameData
                                                    {
                                                        Id = cl.LocationID,
                                                        Name = cl.Description
                                                    });
                        }

                        if (_structureModels.LocationId < 1)
                            LocationId = 0;
                    }
                    catch
                    {
                        _locationTypes = null;
                    }
                }
                return _locationTypes;
            }
        }

        /// <summary>
        /// Retrieve all the alignments
        /// </summary>
        private ObservableCollection<KeyNameData> _alignments;
        public ObservableCollection<KeyNameData> Alignments
        {
            get
            {
                if (_alignments == null && Connection.IsValid)
                {
                    try
                    {
                        _alignments = new ObservableCollection<KeyNameData>();
                        using (var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString))
                        {
                            _alignments.AddRange(from al in dc.Alignments
                                                 where al.SeqTypeID == SequenceTypeId
                                                 orderby al.AlignmentName
                                                 select new KeyNameData
                                                 {
                                                     Id = al.AlnID,
                                                     Name = al.AlignmentName
                                                 });
                        }

                        if (_structureModels.AlignmentId < 0)
                            AlignmentId = _alignments[0].Id;
                    }
                    catch
                    {
                        _alignments = null;
                    }
                }
                return _alignments;
            }
        }

        /// <summary>
        /// Gets/Sets the current step in the creation process (1 or 2)
        /// </summary>
        public int CurrentStep
        {
            get { return _step; }
            set { _step = value; OnPropertyChanged("CurrentStep"); }
        }

        public MTObservableCollection<string> DatabaseNames { get; private set; }
        public ICommand FillDatabaseCommand { get; private set; }
        public ICommand SelectTaxonomyCommand { get; private set; }
        public ICommand NextCommand { get; private set; }
        public ICommand PreviousCommand { get; private set; }

        public SetViewModel()
        {
            _structureModels = new StructureModelSet();
            _shouldSerialize = false;
            _selectSpecificStructureModels = false;
            _selectAllModelsInList = false;
            LocationId = 0;
            SequenceTypeId = 0;
            LimitToStructureModelsWithRendering = true;
            Initialize();
        }

        public SetViewModel(StructureModelSet smsets, bool shouldSerialize, bool selectSpecificStructureModels)
        {
            _structureModels = smsets;
            _shouldSerialize = shouldSerialize;
            _selectSpecificStructureModels = selectSpecificStructureModels;
            _selectAllModelsInList = false;
            Initialize();
        }

        private void Initialize()
        {
            CurrentStep = 1;
            DatabaseNames = new MTObservableCollection<string>();
            FillDatabaseCommand = new DelegatingCommand(() => OnFillDatabases(true));
            SelectTaxonomyCommand = new DelegatingCommand(OnTaxonomySelected);
            PreviousCommand = new DelegatingCommand(OnPreviousStep, () => CurrentStep == 2);
            NextCommand = new DelegatingCommand(OnNextStep, OnCanGotoNextStep);

            if (ParentTaxId != 0)
            {
                ThreadPool.QueueUserWorkItem(o =>
                {
                    for (int i = 0; i < 2; i++)
                    {
                        try
                        {
                            LoadAndFindTaxonomy();
                            break;
                        }
                        catch (SqlException)
                        {
                        }
                        catch
                        {
                            break;
                        }
                    }
                });
            }

            ThreadPool.QueueUserWorkItem(o => OnFillDatabases(false));
        }

        private void OnPreviousStep()
        {
            if (CurrentStep == 2)
                CurrentStep--;
        }

        private bool OnCanGotoNextStep()
        {
            return (CurrentStep == 1)
               ? !string.IsNullOrEmpty(Name) && Data.Connection.IsValid
               : IsValid;
        }

        private void OnNextStep()
        {
            if (CurrentStep == 1)
            {
                CurrentStep++;
            }
            else
                RaiseCloseRequest(true);
        }

        /// <summary>
        /// Fills the database selection combo box
        /// </summary>
        private void OnFillDatabases(bool showError)
        {
            try
            {
                lock (DatabaseNames)
                {
                    if (Connection.IsValid)
                    {
                        DatabaseNames.Clear();
                        foreach (var name in Connection.GetDatabaseList(showError))
                            DatabaseNames.Add(name);

                        if (string.IsNullOrEmpty(Database))
                        {
                            string dbName = DatabaseNames.FirstOrDefault(db => string.Compare(db, "rCAD", true) == 0);
                            if (dbName != null)
                                Database = dbName;
                        }
                    }
                }
            }
            catch
            {
                DatabaseNames.Clear();
                DatabaseNames.Add("rCAD");
                Database = DatabaseNames[0];
            }
        }

        private void LoadAndFindTaxonomy()
        {
            var dc = RcadSecondaryStructureProvider.CreateDbContext(ConnectionString);
            _rootTaxonomy = new TaxonomyEntry(dc, true);

            // Get a list of the taxids leading up to the selected one.
            string sql = @"with cte as (select ParentTaxID from Taxonomy where TaxID={0} " +
                         @"union all select t.ParentTaxID from Taxonomy t join cte tp on t.TaxID = tp.ParentTaxID) " +
                         @"select * from cte";

            var taxIds = dc.ExecuteQuery<int>(sql, ParentTaxId).Reverse().ToList();
            taxIds.RemoveAt(0); // Remove root entry
            taxIds.Add(ParentTaxId);

            TaxonomyEntry current = _rootTaxonomy;

            foreach (int tid in taxIds)
            {
                // Skip root (level 0)
                if (tid <= 1)
                    continue;

                current.IsExpanded = true;
                foreach (var child in current.Children)
                {
                    if (child.Id == tid)
                    {
                        current = child;
                        break;
                    }
                }
            }

            if (current.Id == ParentTaxId)
                current.IsSelected = true;
            else
            {
                _rootTaxonomy = null;
                dc.Dispose();
            }
        }

        /// <summary>
        /// This is used to detect changes in the taxonomy selection.
        /// We could
        /// </summary>
        /// <param name="parameter"></param>
        private void OnTaxonomySelected(object parameter)
        {
            EventParameters ep = (EventParameters)parameter;
            RoutedPropertyChangedEventArgs<object> e = (RoutedPropertyChangedEventArgs<object>)ep.EventArgs;
            TaxonomyEntry tw = (TaxonomyEntry)e.NewValue;
            if (tw != null)
                ParentTaxId = tw.Id;
        }

        private void OnStructureModelSelectedFromList(object sender, EventArgs e)
        {
            if (_sequencesWithStructureModelsSelectionList != null)
            {
                _matchingSequencesWithStructureModels = _structureModels.SelectedSet.Count;
                OnPropertyChanged("IsValid", "MatchingSequenceCount", "HasValidSequenceList");
            }
        }

        /// <summary>
        /// Disposes the view
        /// </summary>
        /// <param name="isDisposing"></param>
        protected override void Dispose(bool isDisposing)
        {
            if (isDisposing)
            {
                if (_rootTaxonomy != null)
                {
                    _rootTaxonomy.Dispose();
                    _rootTaxonomy = null;
                }
            }

            base.Dispose(isDisposing);
        }

        internal StructureModelSet Data
        {
            get { return _structureModels; }
        }

        private RcadConnection Connection
        {
            get { return _structureModels.Connection; }
        }
    }
}
