﻿/* 
* Copyright (c) 2009, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Bio.Data.Interfaces;

namespace Bio.Data.Providers
{
    /// <summary>
    /// This class is an in-memory representation of an RNA sequence.
    /// </summary>
    [DebuggerDisplay("{Id}:{Name}:{ScientificName}")]
    public class SimpleRNASequence : BioExtensiblePropertyBase, IBioSequence
    {
        private int? _firstColumn;

        public string SourceId { get; set; }
        public string Id { get; set; }
        public string Name { get; set; }
        public string ScientificName { get; set; }
        public string TaxonomyId { get; set; }
        
        public int FirstDataColumn 
        {
            get
            {
                if (_firstColumn == null)
                {
                    if (SequenceData.Count == 0)
                        return -1;
                    else
                    {
                        int pos = Enumerable.Range(0, SequenceData.Count).Where(i => SequenceData[i].Type == BioSymbolType.Nucleotide).FirstOrDefault();
                        if (pos == 0 && SequenceData[0].Type != BioSymbolType.Nucleotide)
                            pos = -1;
                        _firstColumn = pos;
                    }
                }
                return _firstColumn.Value;
            }
        }

        public IBioValidator Validator { get; private set; }
        public IList<IBioSymbol> SequenceData { get; private set; }

        public SimpleRNASequence()
        {
            SequenceData = new List<IBioSymbol>();
            Validator = BioValidator.rRNAValidator;
            _firstColumn = null;
        }

        public SimpleRNASequence(IBioValidator validator)
            : this()
        {
            Validator = validator;
        }

        public void AddSymbol(char ch)
        {
            if (Validator.IsValid(ch))
                AddSymbol(new BioSymbol(BioSymbolType.Nucleotide, ch));
            else
                AddSymbol(new BioSymbol(BioSymbolType.Unknown, ch));
        }

        public void AddSymbol(IBioSymbol s)
        {
            SequenceData.Add(s);
        }
    }
}
