﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Diagnostics;
using Bio.Data;
using Bio.Data.Providers.Structure;
using Bio.Views.Structure.Editor.Internal.Rendering;
using Bio.Views.Structure.Editor.ViewModels;

namespace Bio.Views.Structure.Editor.Internal.Editing
{
    public class SecondaryStructureDrawingPropertyContext : StructureElementVisualComponentPropertyContext
    {
        //Directly apply the latest property context state to the visual (this is basically how we support undo, assuming the "latest" property context was rolled back to reflect the undo).
        public override void Apply()
        {
            DrawingViewModel.NucleotideFontFamily = NucleotideFontFamily;
            DrawingViewModel.NucleotideFontStyle = NucleotideFontStyle;
            DrawingViewModel.NucleotideFontWeight = NucleotideFontWeight;
            DrawingViewModel.NucleotideFontSize = NucleotideFontSize;
            DrawingViewModel.NucleotideRenderingStyle = NucleotideRenderingStyle;
            DrawingViewModel.NucleotideSubstitution = NucleotideSubstitution;
            DrawingViewModel.IsNucleotideSubstituted = IsNucleotideSubstituted;
            DrawingViewModel.NucleotideCircleDiameter = NucleotideCircleDiameter;
            DrawingViewModel.NucleotideCircleEdgeThickness = NucleotideCircleEdgeThickness;
            DrawingViewModel.NucleotideCircleFill = NucleotideCircleFill;
            DrawingViewModel.NucleotideCircleIsFilled = NucleotideCircleIsFilled;
            DrawingViewModel.NucleotideCircleStroke = NucleotideCircleStroke;
            DrawingViewModel.BasePairConnectorFill = BasePairConnectorFill;
            DrawingViewModel.BasePairConnectorIsFilled = BasePairConnectorIsFilled;
            DrawingViewModel.BasePairConnectorRenderingStyle = BasePairConnectorRenderingStyle;
            DrawingViewModel.BasePairConnectorSize = BasePairConnectorSize;
            DrawingViewModel.BasePairConnectorStroke = BasePairConnectorStroke;
            DrawingViewModel.BasePairConnectorStrokeThickness = BasePairConnectorStrokeThickness;
        }

        //Save the latest property context state to a new metadata object.
        public override object Save()
        {
            Debug.Assert(DrawingViewModel != null);
            if (DrawingViewModel.IsDirty())
            {
                StructureModelMetadata modelMetadata = new StructureModelMetadata();
                SaveNucleotideRenderingStyle(modelMetadata);
                SaveConnectorRenderingStyle(modelMetadata);
                SaveConnectorProperties(modelMetadata);
                SaveNucleotideProperies(modelMetadata);
                return modelMetadata;
            }
            return null;
        }

        public SecondaryStructureDrawingPropertyContext(SecondaryStructureEditorViewModel drawingViewModel) 
            : base(drawingViewModel.IsComponentVisible)
        {
            Debug.Assert(drawingViewModel != null);
            Debug.Assert(drawingViewModel.CurrentStructure != null);
            DrawingViewModel = drawingViewModel;
            //We set the initial context directly to avoid tracking these as edits.
            CurrentContext.NucleotideFontFamily = DrawingViewModel.NucleotideFontFamily;
            CurrentContext.NucleotideFontStyle = DrawingViewModel.NucleotideFontStyle;
            CurrentContext.NucleotideFontWeight = DrawingViewModel.NucleotideFontWeight;
            CurrentContext.NucleotideFontSize = DrawingViewModel.NucleotideFontSize;
            CurrentContext.NucleotideFill = DrawingViewModel.NucleotideFill;
            CurrentContext.NucleotideRenderingStyle = DrawingViewModel.NucleotideRenderingStyle;
            CurrentContext.NucleotideSubstitution = DrawingViewModel.NucleotideSubstitution;
            CurrentContext.IsNucleotideSubstituted = DrawingViewModel.IsNucleotideSubstituted;
            CurrentContext.NucleotideCircleDiameter = DrawingViewModel.NucleotideCircleDiameter;
            CurrentContext.NucleotideCircleEdgeThickness = DrawingViewModel.NucleotideCircleEdgeThickness;
            CurrentContext.NucleotideCircleFill = DrawingViewModel.NucleotideCircleFill;
            CurrentContext.NucleotideCircleIsFilled = DrawingViewModel.NucleotideCircleIsFilled;
            CurrentContext.NucleotideCircleStroke = DrawingViewModel.NucleotideCircleStroke;
            CurrentContext.BasePairConnectorFill = DrawingViewModel.BasePairConnectorFill;
            CurrentContext.BasePairConnectorIsFilled = DrawingViewModel.BasePairConnectorIsFilled;
            CurrentContext.BasePairConnectorIsVisible = DrawingViewModel.BasePairConnectorIsVisible;
            CurrentContext.BasePairConnectorRenderingStyle = DrawingViewModel.BasePairConnectorRenderingStyle;
            CurrentContext.BasePairConnectorSize = DrawingViewModel.BasePairConnectorSize;
            CurrentContext.BasePairConnectorStroke = DrawingViewModel.BasePairConnectorStroke;
            CurrentContext.BasePairConnectorStrokeThickness = DrawingViewModel.BasePairConnectorStrokeThickness;
        }

        private SecondaryStructureEditorViewModel DrawingViewModel { get; set; }

        private void SaveNucleotideRenderingStyle(StructureModelMetadata metadata)
        {
            if (NucleotideRenderingStyle.HasValue)
            {
                switch (NucleotideRenderingStyle.Value)
                {
                    case NucleotideRenderingMode.Circle:
                        metadata.NucleotideRenderingMode = NucleotideMetadata.RenderingMode.Circle;
                        break;
                    case NucleotideRenderingMode.Text:
                        metadata.NucleotideRenderingMode = NucleotideMetadata.RenderingMode.Text;
                        break;
                    default:
                        metadata.NucleotideRenderingMode = null;
                        break;
                };
            }
            else { metadata.NucleotideRenderingMode = null; }
        }

        private void SaveConnectorRenderingStyle(StructureModelMetadata metadata)
        {
            if (BasePairConnectorRenderingStyle.HasValue)
            {
                switch (BasePairConnectorRenderingStyle.Value)
                {
                    case BasePairConnectorRenderingMode.Circle:
                        metadata.ConnectorStyle = BasePairMetadata.ConnectorRenderingMode.Circle;
                        break;
                    case BasePairConnectorRenderingMode.Line:
                        metadata.ConnectorStyle = BasePairMetadata.ConnectorRenderingMode.Line;
                        break;
                    case BasePairConnectorRenderingMode.Rectangle:
                        metadata.ConnectorStyle = BasePairMetadata.ConnectorRenderingMode.Rectangle;
                        break;
                    default:
                        metadata.ConnectorStyle = null;
                        break;
                };
            }
            else { metadata.ConnectorStyle = null; }
        }

        private void SaveConnectorProperties(StructureModelMetadata metadata)
        {
            metadata.ConnectorVisible = BasePairConnectorIsVisible;
            if (ContextHasConnectorRenderingPropertiesSet())
            {
                if(metadata.ConnectorRendering == null)
                    metadata.ConnectorRendering = new RNASecondaryStructureBasePairConnector();

                metadata.ConnectorRendering.FillColor = (BasePairConnectorFill != null) ? BasePairConnectorFill.ToString() : null;
                metadata.ConnectorRendering.Filled = BasePairConnectorIsFilled;
                metadata.ConnectorRendering.Size = (BasePairConnectorSize.HasValue) ? BasePairConnectorSize.Value : 0.0;
                metadata.ConnectorRendering.StrokeColor = (BasePairConnectorStroke != null) ? BasePairConnectorStroke.ToString() : null;
                metadata.ConnectorRendering.EdgeThickness = (BasePairConnectorStrokeThickness.HasValue) ? BasePairConnectorStrokeThickness.Value : 0.0;
            }
            else { metadata.ConnectorRendering = null; }
        }

        private void SaveNucleotideProperies(StructureModelMetadata metadata)
        {
            //Check if we have nucleotide rendering properties set.
            if (ContextHasCircleRenderingPropertiesSet() ||
                ContextHasTextRenderingPropertiesSet())
            {
                if (ContextHasTextRenderingPropertiesSet())
                {
                    if (metadata.NucleotideTextRendering == null)
                        metadata.NucleotideTextRendering = new RNASecondaryStructureNucleotideTextShape();

                    metadata.NucleotideTextRendering.Font = (NucleotideFontFamily != null) ? NucleotideFontFamily.Source : null;
                    metadata.NucleotideTextRendering.FontStyle = (NucleotideFontStyle.HasValue) ? NucleotideFontStyle.Value.ToString() : null;
                    metadata.NucleotideTextRendering.FontWeight = (NucleotideFontWeight.HasValue) ? NucleotideFontWeight.Value.ToString() : null;
                    metadata.NucleotideTextRendering.FontSize = (NucleotideFontSize.HasValue) ? NucleotideFontSize.Value : 0.0;
                    metadata.NucleotideTextRendering.Color = (NucleotideFill != null) ? NucleotideFill.ToString() : null;
                }
                else
                {
                    metadata.NucleotideTextRendering = null;
                }

                if (ContextHasCircleRenderingPropertiesSet())
                {
                    if (metadata.NucleotideCircleRendering == null)
                        metadata.NucleotideCircleRendering = new RNASecondaryStructureNucleotideCircleShape();

                    metadata.NucleotideCircleRendering.EdgeThickness = (NucleotideCircleEdgeThickness.HasValue && NucleotideCircleEdgeThickness.Value > 0.0) ? NucleotideCircleEdgeThickness.Value : 0.0;
                    metadata.NucleotideCircleRendering.Diameter = (NucleotideCircleDiameter.HasValue && NucleotideCircleDiameter.Value > 0.0) ? NucleotideCircleDiameter.Value : 0.0;
                    metadata.NucleotideCircleRendering.Filled = NucleotideCircleIsFilled;
                    metadata.NucleotideCircleRendering.FillColor = (NucleotideCircleFill != null) ? NucleotideCircleFill.ToString() : null;
                    metadata.NucleotideCircleRendering.EdgeColor = (NucleotideCircleStroke != null) ? NucleotideCircleStroke.ToString() : null;
                }
                else
                {
                    metadata.NucleotideCircleRendering = null;
                }
            }
        }
    }
}
