﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using Bio.Data.Interfaces;
using Bio.Views.Structure.Editor.Internal.Editing;
using Bio.Views.Structure.Editor.Internal.Rendering;
using Bio.Views.Structure.Editor.Internal.Rendering.Interfaces;
using Bio.Views.Structure.Editor.Internal.Rendering.Visuals;

namespace Bio.Views.Structure.Editor.Internal.Selection
{
    public class DomainSelectionGroup : StructureElementSelectionGroupBase
    {
        public override int ID { get { return _contextualHelix.FivePrimeStartNucleotideIndex; } }
        public override IStructureElementVisualComponent Visual { get { return _rangeVisual; } }

        public override void MakeElementSelectable()
        {
            _contextualHelix.MakeSelectable();
            //We have to make sure all nt are selectable, unfortunately, this could be expensive in a large diagram
            foreach (var ntVisual in _rangeVisual.AllNucleotideVisuals.Where(ntVisual => ntVisual.NucleotideIndex >= _contextualHelix.FivePrimeStartNucleotideIndex &&
                ntVisual.NucleotideIndex <= _contextualHelix.ThreePrimeEndNucleotideIndex))
                ntVisual.MakeSelectable();
        }

        public override void MakeElementUnselectable()
        {
            _contextualHelix.MakeUnselectable();
            //We have to make sure all nt unselectable when we toggle off, unfortunately, this could be expensive in a large diagram
            foreach (var ntVisual in _rangeVisual.AllNucleotideVisuals.Where(ntVisual => ntVisual.NucleotideIndex >= _contextualHelix.FivePrimeStartNucleotideIndex &&
                ntVisual.NucleotideIndex <= _contextualHelix.ThreePrimeEndNucleotideIndex))
                ntVisual.MakeUnselectable();
        }

        public override void SelectElement()
        {
            _rangeVisual.SetFirstSelection(_contextualHelix.FivePrimeStartNucleotideIndex);
            _rangeVisual.SetSecondSelection(_contextualHelix.ThreePrimeEndNucleotideIndex);
        }

        public override void DeselectElement()
        {
            _rangeVisual.ClearSelection();
        }

        public override void MoveBy(Point delta)
        {
            if (_rangeVisual.SelectedNucleotideVisuals.Count() > 0 || _rangeVisual.SelectedBasePairVisuals.Count() > 0)
            {
                //1. Aggregate into one list
                List<int> ntInRange = new List<int>();
                foreach (var ntVisual in _rangeVisual.SelectedNucleotideVisuals)
                    ntInRange.Add(ntVisual.NucleotideIndex);
                foreach (var bpVisual in _rangeVisual.SelectedBasePairVisuals)
                {
                    ntInRange.Add(bpVisual.Data.FivePrimeIndex);
                    ntInRange.Add(bpVisual.Data.ThreePrimeIndex);
                }
                foreach (var ntIdx in ntInRange)
                {
                    Manager.MoveNucleotideBy(ntIdx, delta, true, true, true);
                }
                Manager.SaveStructureElementDrawingOperation(ntInRange, delta);
            }
        }

        public override void EditProperties(StructureElementPropertyChangeSet changeSet, List<IBioSymbol> filter)
        {
            if (_rangeVisual.SelectedNucleotideVisuals.Count() > 0 || _rangeVisual.SelectedBasePairVisuals.Count() > 0)
            {
                _rangeVisual.ApplyPropertyEditSet(changeSet, false, filter);
            }
        }

        public override void DeleteSelection()
        {
            //We only need to execute if we have a completed selection.
            if (_rangeVisual.SelectedBasePairVisuals.Count() > 0)
            {
                List<int> localSelectedBasePairs = new List<int>();
                foreach (var sbp in _rangeVisual.SelectedBasePairVisuals)
                    localSelectedBasePairs.Add(sbp.Data.FivePrimeIndex);
                Manager.RemoveBasePairs(localSelectedBasePairs);
            }
        }

        public DomainSelectionGroup(StructureDrawingManager manager, HelixStructureElementVisualComponent contextHelix)
            : base(manager, contextHelix)
        {
            Debug.Assert(manager.FullModelRangeSelectionVisual != null);
            Debug.Assert(contextHelix != null);
            _contextualHelix = contextHelix;
            _rangeVisual = manager.FullModelRangeSelectionVisual;
        }

        private HelixStructureElementVisualComponent _contextualHelix;
        private RangeStructureElementVisualComponent _rangeVisual;
    }
}
