﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Windows;
using System.Windows.Input;
using Bio.Views.Structure.Editor.Internal.Selection;

namespace Bio.Views.Structure.Editor.Internal.Tools
{
    public class StructureElementDrawingTool : SecondaryStructureDrawingTool
    {
        public StructureElementDrawingTool()
        {
            Icon = "E";
            Message = string.Empty;
            ToolTip = "Drag Edit Tool; Select a Structure Element and drag it to a new location";
        }

        public override void OnSelectionModeChanged()
        {
            if(IsActive)
                SetMessage();
        }

        protected override void DrawingMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            if (Drawing.TryStartDragEdit(clickLocation))
                GoToNextState();
            else
            {
                Message = string.Format("Error: For Ranges and Domains, use the Selection Tool to make the Selection and then switch back to the Drag Edit Tool to Edit");
            }
        }

        protected override void DrawingMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            Point delta;
            if (Drawing.TryEndDragEdit(clickLocation, out delta))
                GoToNextState();
        }

        protected override void DrawingMouseMove(object sender, MouseEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            Point delta;
            if (Drawing.TryUpdateDragEdit(clickLocation, out delta))
            {
                Delta = delta;
                SetMessage();
            }

        }

        protected override void DrawingMouseLeave(object sender, MouseEventArgs e)
        {
            Point clickLocation = e.GetPosition(Drawing);
            Point delta;
            if (Drawing.TryEndDragEdit(clickLocation, out delta))
                GoToNextState();
        }

        protected override void OnActiveStateChanged()
        {
            if (!IsActive)
                Message = null;
            else
            {
                State = DrawingState.Waiting;
                SetMessage();
            }
        }

        private DrawingState State { get; set; }
        private enum DrawingState { Waiting, Dragging, Complete };
        private static string DRAWING_TOOL_TITLE = "Drag Edit Tool";
        private Point Delta { get; set; }
        private Point LastDelta { get; set; }

        private void SetMessage()
        {
            switch (State)
            {
                case DrawingState.Waiting:
                    if (Drawing.SelectionMode == StructureElementSelectionMode.Range)
                    {
                        Message = string.Format("{0}: For Ranges, use the Selection Tool to make the Selection, then switch back to the Drag Edit Tool", DRAWING_TOOL_TITLE);
                    }
                    else if (Drawing.SelectionMode == StructureElementSelectionMode.None)
                    {
                        Message = string.Format("{0}: Must set the Selection Mode to something other than None to use this Tool", DRAWING_TOOL_TITLE);
                    }
                    else
                    {
                        Message = DRAWING_TOOL_TITLE;
                    }
                    break;
                case DrawingState.Dragging:
                    Message = string.Format("{0}: Move Delta (x,y): ({1:F},{2:F})", DRAWING_TOOL_TITLE, Delta.X, Delta.Y);
                    break;
                case DrawingState.Complete:
                    if (Drawing.SelectionMode == StructureElementSelectionMode.Range)
                    {
                        Message = string.Format("{0}: Last Move (x,y): ({1:F},{2:F}), switch back to the Selection Tool to reselect the Range before starting another Drag Edit.", DRAWING_TOOL_TITLE, LastDelta.X, LastDelta.Y);
                    }
                    else
                    {
                        Message = string.Format("{0}: Last Move (x,y): ({1:F},{2:F})", DRAWING_TOOL_TITLE, LastDelta.X, LastDelta.Y);
                    }
                    break;
                default:
                    Message = DRAWING_TOOL_TITLE;
                    break;
            };
        }

        private void GoToNextState()
        {
            if (State == DrawingState.Waiting)
                State = DrawingState.Dragging;
            else if (State == DrawingState.Dragging)
            {
                State = DrawingState.Complete;
                LastDelta = Delta;
                Delta = new Point();
            }
            else if (State == DrawingState.Complete)
                State = DrawingState.Dragging;
            SetMessage();
        }
    }
}
