﻿/* 
* Copyright (c) 2010, The University of Texas at Austin
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification, 
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, 
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice, 
* this list of conditions and the following disclaimer in the documentation and/or other materials 
* provided with the distribution.
*
* Neither the name of The University of Texas at Austin nor the names of its contributors may be 
* used to endorse or promote products derived from this software without specific prior written 
* permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows.Input;
using Bio.Views.Structure.Editor.Internal.Data;
using Bio.Views.Structure.Editor.Views;
using JulMar.Windows;
using JulMar.Windows.Interfaces;
using JulMar.Windows.Mvvm;

namespace Bio.Views.Structure.Editor.ViewModels
{
    public class BasePairVisualRenderingDefaultsViewModel : ViewModel
    {
        public BasePairVisualRenderingOptionsViewModel SelectedOptionsSet
        {
            get { return _selectedOptionsSet; }
            set
            {
                _selectedOptionsSet = value;
                OnPropertyChanged("SelectedOptionsSet");
            }
        }

        public MTObservableCollection<BasePairVisualRenderingOptionsViewModel> OptionsSets
        {
            get { return _allOptions; }
            private set
            {
                _allOptions = value;
                OnPropertyChanged("OptionsSets");
            }
        }

        public BasePairVisualRenderingOptionsDataCollection RuntimeOptionsSets
        {
            get;
            private set;
        }

        public ICommand AddNewNucleotideRenderingDefaultsOptionSetCommand { get; private set; }
        public ICommand ResetRuntimeDefaultBasePairRenderingOptionsToBuiltInDefaultsCommand { get; private set; }

        public BasePairVisualRenderingDefaultsViewModel(BasePairVisualRenderingOptionsDataCollection runtimeOptionsSets, BasePairVisualRenderingOptionsDataCollection defaultOptionsSet)
        {
            Debug.Assert(defaultOptionsSet != null);
            Debug.Assert(runtimeOptionsSets != null);
            _defaultOptionsSet = defaultOptionsSet;
            BasePairVisualRenderingOptionsDataCollection localRuntimeOptionsSets = new BasePairVisualRenderingOptionsDataCollection();
            foreach (var optionSet in runtimeOptionsSets)
                localRuntimeOptionsSets.Add(new BasePairVisualRenderingOptionsData(optionSet));
            RuntimeOptionsSets = localRuntimeOptionsSets;
            Initialize();   
        }

        private void Initialize()
        {
            ResetRuntimeDefaultBasePairRenderingOptionsToBuiltInDefaultsCommand = new DelegatingCommand(OnResetRuntimeDefaultBasePairOptionSets);
            AddNewNucleotideRenderingDefaultsOptionSetCommand = new DelegatingCommand(OnAddNewBasePairDefaultOptionSet);
            OptionsSets = new MTObservableCollection<BasePairVisualRenderingOptionsViewModel>();
            foreach (var optionSet in RuntimeOptionsSets)
            {
                OptionsSets.Add(new BasePairVisualRenderingOptionsViewModel(optionSet));
            }

            SelectedOptionsSet = OptionsSets[0];
            _validBasePairs = new List<string>()
            {
                "A-U",
                "U-A",
                "G-C",
                "C-G",
                "G-U",
                "U-G",
                "A-A",
                "A-C",
                "A-G",
                "U-U",
                "U-C",
                "G-G",
                "G-A",
                "C-C",
                "C-A",
                "C-U"
            };
        }

        private void OnAddNewBasePairDefaultOptionSet()
        {
            //We will have to put up a simple UI to solicit a label
            SelectBasePairForRenderingDefaultsViewModel vm = new SelectBasePairForRenderingDefaultsViewModel(_validBasePairs);
            IUIVisualizer uiVisualizer = ViewModel.ServiceProvider.Resolve<IUIVisualizer>();
            if (uiVisualizer != null)
            {
                bool? result = uiVisualizer.ShowDialog(typeof(SelectBasePairForRenderingDefaultsEditor).ToString(), vm);
                if (result != null && result.Value && vm.SelectedValue != null)
                {
                    var existingOptionSet = OptionsSets.Where(set => set.Label.Equals(vm.SelectedValue)).FirstOrDefault();
                    if (existingOptionSet != null)
                    {
                        IErrorVisualizer errorVisualizer = Resolve<IErrorVisualizer>();
                        if (errorVisualizer != null)
                        {
                            errorVisualizer.Show("Error", string.Format("Error: Already have an Options Set for Base Pair: {0}", vm.SelectedValue));
                        }
                    }
                    else
                    {
                        BasePairVisualRenderingOptionsData newOptionsSet = BasePairVisualRenderingOptionsData.CreateDefault(vm.SelectedValue);
                        BasePairVisualRenderingOptionsViewModel newOptionsSetVM = new BasePairVisualRenderingOptionsViewModel(newOptionsSet);
                        OptionsSets.Add(newOptionsSetVM);
                        RuntimeOptionsSets.Add(newOptionsSet);
                        SelectedOptionsSet = newOptionsSetVM; //Make this new set selected.
                    }
                }
            }
        }

        private void OnResetRuntimeDefaultBasePairOptionSets()
        {
            //We'll just replace the runtime with the default.
            RuntimeOptionsSets.Clear();
            foreach (var defaultOptionSet in _defaultOptionsSet)
                RuntimeOptionsSets.Add(new BasePairVisualRenderingOptionsData(defaultOptionSet));
            OptionsSets.Clear();
            foreach (var defaultOptionSet in _defaultOptionsSet)
                OptionsSets.Add(new BasePairVisualRenderingOptionsViewModel(defaultOptionSet));
            SelectedOptionsSet = OptionsSets[0];
        }

        private MTObservableCollection<BasePairVisualRenderingOptionsViewModel> _allOptions;
        private List<string> _validBasePairs;
        private BasePairVisualRenderingOptionsViewModel _selectedOptionsSet;
        private BasePairVisualRenderingOptionsDataCollection _defaultOptionsSet;
    }
}
